
import numpy
import pickle
import scipy
import logging 
from apgl.util.Util import Util
from apgl.util.Parameter import Parameter
from apgl.graph.AbstractSingleGraph import AbstractSingleGraph
from apgl.graph.VertexList import VertexList
from apgl.predictors import * 

class AbstractMatrixGraph(AbstractSingleGraph):
    """
    An abstract graph object with single edges between vertices. The edge labels
    are stored in a weight matrix, and vertices are stored in a VertexList 
    """

    def getWeightMatrix(self):
        """
        Returns a numpy array of the weight matrix of this graph.
        """
        Util.abstract()

    def laplacianMatrix(self, outDegree=True):
        """
        Return the Laplacian matrix of this graph, which is defined a L_{ii} = deg(i)
        L_{ij} = -1 if an edge between i and j, otherwie L_{ij} = 0 . For a directed
        graph the uses the out-degree.
        """
        numVertices = self.getNumVertices()
        A = self.adjacencyMatrix()

        if outDegree:
            return -A + numpy.diag(numpy.dot(A, numpy.ones((numVertices, 1))).ravel())
        else:
            return -A + numpy.diag(numpy.dot(numpy.ones((1, numVertices)), A).ravel())

    def getAllVertexIds(self):
        """
        Returns a list of all the vertex IDs of this graph. 
        """
        return range(0, self.vList.getNumVertices())

    def getVertexList(self):
        """
        Returns the VertexList object of this graph. 
        """
        return self.vList

    def getVertex(self, vertexIndex):
        """
        Returns the vertex associated with the given vertex index.

        :param vertexIndex: the index of the vertex.
        :type vertexIndex: :class:`int`
        """
        #The vertexlist should check the parameter (no point doing it twice) 
        return self.vList.getVertex(vertexIndex)

    def setVertex(self, vertexIndex, vertex):
        """
        Set the vertex with given index to a particular value. 

        :param vertexIndex: the index of the vertex.
        :type vertexIndex: :class:`int`

        :param vertex: the value of the vertex.
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())
        self.vList.setVertex(vertexIndex, vertex)

    def getNumVertices(self):
        """
        Return the number of vertices in this graph. 
        """
        return self.vList.getNumVertices()

    def getNumVertexFeatures(self):
        """
        Return the number of vertex features.
        This method is depricated.
        """
        return self.vList.getNumFeatures()

    def addEdge(self, vertexIndex1, vertexIndex2, edge=1):
        """
        Add a non-zero edge between two vertices.

        :param vertexIndex1: The index of the first vertex.
        :type vertexIndex1: :class:`int`

        :param vertexIndex2: The index of the second vertex.
        :type vertexIndex2: :class:`int`

        :param edge: The value of the edge.
        :type edge: :class:`float`
        """
        Parameter.checkIndex(vertexIndex1, 0, self.vList.getNumVertices())
        Parameter.checkIndex(vertexIndex2, 0, self.vList.getNumVertices())

        if edge == 0 or edge == float('inf'):
            raise ValueError("Cannot add a zero or infinite edge")

        self.W[vertexIndex1, vertexIndex2] = edge
        if self.undirected:
            self.W[vertexIndex2, vertexIndex1] = edge

    def addEdges(self, edgeIndexArray, edgeValues=[]):
        """
        Takes a numpy array of edge index pairs, and edge values and adds them
        to this graph. The array is 2 dimensional such that each row is a pair
        of edge indices.

        :param edgeIndexArray: The array of edge indices with each being a pair of indices.
        :type edgeIndexArray: :class:`numpy.ndarray`

        :param edgeValues: The list of edge values
        :type edgeValues: :class:`list`
        """
        if (edgeIndexArray < 0).any() or (edgeIndexArray >= self.vList.getNumVertices()).any():
            raise ValueError("Invalid indices for edges.")
        

        if edgeValues != []:
            if (edgeValues == 0).any():
                raise ValueError("Invalid entry, found zero edge value(s): " + str(numpy.nonzero(edgeValues==0)[0]))

            if self.undirected:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = edgeValues[i]
                    self.W[edgeIndexArray[i, 1], edgeIndexArray[i, 0]] = edgeValues[i]
            else:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = edgeValues[i]
        else:
            if self.undirected:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = 1
                    self.W[edgeIndexArray[i, 1], edgeIndexArray[i, 0]] = 1
            else:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = 1

    def getEdge(self, vertexIndex1, vertexIndex2):
        """
        Get the value of an edge.

        :param vertexIndex1: The index of the first vertex.
        :type vertexIndex1: :class:`int`

        :param vertexIndex2: The index of the second vertex.
        :type vertexIndex2: :class:`int`
        """
        Parameter.checkIndex(vertexIndex1, 0, self.vList.getNumVertices())
        Parameter.checkIndex(vertexIndex2, 0, self.vList.getNumVertices())

        if self.W[vertexIndex1, vertexIndex2]==0:
            return None
        else:
            return self.W[vertexIndex1, vertexIndex2]

    def getEdgeValues(self, edgeArray):
        """
        Take an array of n x 2 of vertex indices and return the corresponding edge
        values.
        """
        edgeValues = numpy.zeros(edgeArray.shape[0])

        for i in range(edgeValues.shape[0]):
            edgeValues[i] = self.getEdge(edgeArray[i, 0], edgeArray[i, 1])

        return edgeValues 

    def removeEdge(self, vertexIndex1, vertexIndex2):
        """
        Remove an edge between two vertices.

        :param vertexIndex1: The index of the first vertex.
        :type vertexIndex1: :class:`int`

        :param vertexIndex2: The index of the second vertex.
        :type vertexIndex2: :class:`int`
        """
        Parameter.checkIndex(vertexIndex1, 0, self.vList.getNumVertices())
        Parameter.checkIndex(vertexIndex2, 0, self.vList.getNumVertices())
        self.W[vertexIndex1, vertexIndex2] = 0
        if self.undirected:
            self.W[vertexIndex2, vertexIndex1] = 0

    def isUndirected(self):
        """
        Returns true if this graph is undirected otherwise false. 
        """
        return self.undirected

    def getAllEdges(self):
        """
        For an undirected graph will get v1>=v2. Return a matrix with 2 columns
        """
        (rows, cols) = numpy.nonzero(self.W)
        edges = numpy.c_[rows, cols]

        if self.undirected and edges.shape[0] != 0:
            edges = edges[edges[:,0] >= edges[:,1], :]

        return edges

    def removeAllEdges(self):
        """
        Removes all edges from this graph. 
        """
        self.W = self.W*0

    def __str__(self):
        output = str(type(self)) + ": "
        output = output + "no. vertices " + str(self.getNumVertices()) + ", no. edges, " + str(self.getNumEdges()) + " undirected: " + str(self.undirected)
        return output


    def diameter(self, useWeights=False, P=None):
        """
        Finds the diameter of a graph i.e. the longest shortest path. If useWeights
        is True then the weights in the adjacency matrix are used if P is not
        provided. 

        :param useWeights: Whether to use edge weights to compute a diameter. 
        :type useWeights: :class:`bool`

        :param P: An optional nxn matrix whose ijth entry is the shortest path
        from i to j.
        :type P: :class:`ndarray`
        """
        Parameter.checkBoolean(useWeights)
        if P!=None and (type(P) != numpy.ndarray or P.shape != self.W.shape):
            raise ValueError("P must be array of same size as weight matrix of graph")
        
        if self.getNumEdges() == 0: 
            return 0 

        if P == None:
            P = self.floydWarshall(useWeights)
        else:
            P = P.copy()

        if useWeights == False:
            return int(numpy.max(P[P!=float('inf')]))
        else:
            return float(numpy.max(P[P!=float('inf')]))

    def effectiveDiameter(self, q, P=None):
        """
        The effective diameter is the minimum d such that for a fraction q of
        reachable node pairs, the path length is at most d. This is more rubust
        than the standard diameter method. One can optionally pass in a matrix
        P whose ijth entry is the shortest path from i to j. 

        :param q: The fraction of node pairs to consider.
        :type q: :class:`float`

        :param P: An optional nxn matrix whose ijth entry is the shortest path
        from i to j.
        :type P: :class:`ndarray`
        """
        Parameter.checkFloat(q, 0.0, 1.0)
        if P!=None and (type(P) != numpy.ndarray or P.shape != self.W.shape):
            raise ValueError("P must be array of same size as weight matrix of graph")

        if self.getNumEdges() == 0:
            return 0

        #Paths from a vertex to itself are ignored 
        if P == None:
            P = self.floydWarshall(False)
        else:
            P = P.copy()
        
        P[numpy.diag_indices(P.shape[0])] = float('inf')

        paths = numpy.sort(P[P!=float('inf')])
        
        if paths.shape[0] != 0:
            ind = numpy.floor((paths.shape[0]-1)*q)
            return int(paths[ind])
        else:
            return 0.0 
        

    def floydWarshall(self, useWeights=True):
        """
        Use the Floyd-Warshall algorithm to find the shortest path between all pairs
        of vertices. If useWeights is true, then the weights are used to compute the
        path, otherwise adjacencies are used. Note that the shortest path of a
        vertex to itself is always zero. 

        :param useWeights: Whether to use the edge weight to compute path cost. 
        :type useWeights: :class:`bool`
        """
        numVertices = self.vList.getNumVertices()
        
        if useWeights:
            P = self.getWeightMatrix().copy()
        else:
            P = self.adjacencyMatrix().copy()

        P = P.astype(numpy.float64)
        P[P==0] = float('inf')
        
        j = numpy.ones(numVertices)

        for k in range(0, numVertices):
            P2 = numpy.outer(P[:, k], j) + numpy.outer(j, P[k, :])
            P = numpy.minimum(P, P2)

        P[numpy.diag_indices(P.shape[0])] = 0 

        return P

    def maxProductPaths(self):
        """
        Find the maximum product paths between all pairs of vertices using
        a modified version of the Floyd-Warshall algorithm.
        """
        numVertices = self.vList.getNumVertices()
        P = self.getWeightMatrix().copy()
        stepSize = 100

        for k in range(0, numVertices):
            Util.printIteration(k, stepSize, numVertices)
            P2 = numpy.outer(P[:, k], P[k, :])
            P = numpy.maximum(P, P2)

        return P

    def geodesicDistance(self, P=None):
        """
        Compute the mean geodesic distance for a graph. This is denoted for an 
        undirected graph by 1/(1/2 n(n+1)) \sum_{i<=j} d_ij where d_ij is the
        shortest path length between i and j. Note that if i and j are not connected
        we assume a path length of 0. If the graph is directed then the geodesic
        distance is 1/(n(n+1)) \sum_{i, j} d_ij. 

        :param P: An optional nxn matrix whose ijth entry is the shortest path
        from i to j.
        :type P: :class:`ndarray`
        """
        if P!=None and (type(P) != numpy.ndarray or P.shape != self.W.shape):
            raise ValueError("P must be array of same size as weight matrix of graph")

        if self.getNumVertices() == 0:
            return 0
        
        if P == None:
            P = self.floydWarshall(True)
        else:
            P = P.copy()

        n = self.getNumVertices()
        P[P==numpy.inf] = 0
        P[numpy.diag_indices(P.shape[0])] = 0.0

        distanceSum = numpy.sum(P)
        if self.isUndirected():
            return distanceSum/(n*(n+1))
        else:
            return distanceSum/(n**2)

    def harmonicGeodesicDistance(self, P=None):
        """
        Compute the inverse "harmonic mean" geodesic distance for a graph. This is
        denoted by 1/(1/2 n(n+1)) \sum_{i<=j} d_ij^-1 where d_ij is the shortest
        path length between i and j for an undirected graph. The distance from a
        node to itself is infinite. For a directed graph, the distance is
        1/n^2 \sum_{i,j} d_ij^-1.

        :param P: An optional nxn matrix whose ijth entry is the shortest path
        from i to j.
        :type P: :class:`ndarray`
        """
        if P!=None and (type(P) != numpy.ndarray or P.shape != self.W.shape):
            raise ValueError("P must be array of same size as weight matrix of graph")

        if self.getNumVertices() == 0:
            return 0

        if P == None:
            P = self.floydWarshall(True)
        else:
            P = P.copy()

        n = self.getNumVertices()
        P = self.floydWarshall()
        P = 1/(P + numpy.diag(numpy.ones(n)*numpy.inf))

        if self.isUndirected(): 
            distanceSum = numpy.sum(numpy.triu(P))
            return (2*distanceSum)/(n*(n+1))
        else: 
            distanceSum = numpy.sum(P)
            return distanceSum/n**2

    def adjacencyMatrix(self):
        """
        Return the adjacency matrix in dense format. Warning: should not be used
        unless sufficient memory is available to store the dense matrix.
        """
        W2 = self.getWeightMatrix().copy()
        W2[W2.nonzero()] = 1
        return W2

    def maybeIsomorphicWith(self, graph):
        """
        Returns false if graph is definitely not isomorphic with the current graph,
        however a True may mean the graphs are not isomorphic. Makes a comparison
        with the eigenvalues of the Laplacian matrices. 
        """
        L1 = self.laplacianMatrix()
        L2 = graph.laplacianMatrix()

        sigma1, V1 = numpy.linalg.eig(L1)
        sigma2, V2 = numpy.linalg.eig(L2)

        sigma1 = numpy.sort(sigma1)
        sigma2 = numpy.sort(sigma2)

        tol = 10**-6

        return numpy.linalg.norm(sigma1 - sigma2) <= tol

    def complement(self):
        """
        Returns a graph with identical vertices (same reference) to the current one, but with the
        complement of the set of edges. Edges that do not exist have weight 1.
        """

        Util.abstract()

    def vertexLabelPairs(self, edgeArray):
        """
        Create an array of pairs of vertex labels for each edge in the graph. Returns
        a tuple of this array and a vector of corresponding edge weights, for the
        indices of the full edge index array as given by getAllEdges.

        This method is deprecated.

        :param edgeArray: A numpy array with 2 columns, and values corresponding to vertex indices.
        :type edgeArray: :class:`numpy.ndarray`

        """
        #Parameter.checkList(edgeArray, Parameter.checkIndex, (0, self.getNumVertices()))

        numFeatures = self.getVertexList().getNumFeatures()
        X = numpy.zeros((edgeArray.shape[0], numFeatures*2))

        for i in range(edgeArray.shape[0]):
            X[i, 0:numFeatures] = self.getVertex(edgeArray[i, 0])
            X[i, numFeatures:numFeatures*2] = self.getVertex(edgeArray[i, 1])

        return X 

    def save(self, filename):
        """
        Save the graph object to the corresponding file. The python pickle package
        is used for serialisation, and data is stored in a binary format.
        
        :param filename: The name of the file to save.
        :type filename: :class:`str`
        """
        Parameter.checkClass(filename, str) 

        file = open(filename, 'wb')
        pickle.dump(self, file)
        file.close()

        logging.info("Dumped file into " + filename)

    @staticmethod
    def load(filename):
        """
        Load the graph object from the corresponding file. The python pickle package
        is used for serialisation, and data is loaded in a binary format as created
        using save().

        :param filename: The name of the file to load.
        :type filename: :class:`str`
        """
        Parameter.checkClass(filename, str)

        file = open(filename, 'rb')
        graph = pickle.load(file)
        file.close()

        logging.info("Loaded pickled graph: " + filename + " with " + str(graph.getNumVertices()) + " vertices and " + str(graph.getNumEdges()) + " edges")

        return graph

    def setVertexList(self, vList):
        """
        Assign a new VertexList object to this graph. The number of vertices in the
        VertexList must be the same as in the graph. 
        """
        Parameter.checkClass(vList, VertexList)

        if vList.getNumVertices() != self.vList.getNumVertices():
            raise ValueError("Can only set to a VertexList with same number of vertices.")

        self.vList = vList 

    def setWeightMatrix(self, W):
        """
        Set the weight matrix of this graph. Requires as input an ndarray with the
        same dimensions as the current weight matrix. Edges are represented by
        non-zero edges.

        :param W: The name of the file to load.
        :type W: :class:`ndarray`
        """
        Parameter.checkClass(W, numpy.ndarray)

        if W.shape != (self.vList.getNumVertices(), self.vList.getNumVertices()):
            raise ValueError("Weight matrix has wrong shape : " + str(W.shape))

        if self.undirected and (W != W.T).any():
            raise ValueError("Weight matrix of undirected graph must be symmetric")


        self.W = W 

    def outDegreeSequence(self):
        """
        Return a vector of the (out)degree for each vertex.
        """
        Util.abstract()

    def inDegreeSequence(self):
        """
        Return a vector of the (in)degree for each vertex.
        """
        Util.abstract()

    def degreeDistribution(self):
        """
        Return a vector of (out)degree distributions. The ith element of the vector
        corresponds to the frequency of degree i.
        """
        degrees = self.outDegreeSequence()

        binCounts = numpy.bincount(degrees)
        return binCounts

    def inDegreeDistribution(self):
        """
        Returns a vector of in-degree distributions. The ith element of the vector
        corresponds to the frequency of degree i.
        """
        degrees = self.inDegreeSequence()

        binCounts = numpy.bincount(degrees)
        return binCounts

    def fitPowerLaw(self):
        """
        Fits the out-degree probabilities of this graph using the power law
        p_d ~ d^-gamma. The value of xmin is the point to start taking examples. 
        """
        logging.warn("Use with caution: fitPowerLaw may not be stable")
        degreeSeq = self.outDegreeSequence()
        degreeMax = numpy.max(degreeSeq)
        xmins = numpy.arange(1, numpy.minimum(20, degreeMax))
        ks, alpha, xmin = Util.fitDiscretePowerLaw(degreeSeq, xmins)

        return alpha, ks, xmin

    def hopCount(self):
        """
        Returns an array such that the ith element is the number of pairs of
        vertices reachable within i hops. This includes self pairs, and all
        other pairs are counted twice in the undirected case otherwise once.
        """
        p = self.floydWarshall(False).ravel()
        p = p[numpy.logical_not(numpy.isinf(p))]
        p = numpy.array(p, numpy.int32)
        hopCount = numpy.bincount(p)

        return numpy.cumsum(hopCount)

    def triangleSequence(self):
        """
        Computes the number of triangles each vertex participates in using the
        diagonal of the adjcancy matrix. In an undirected graph, a each triangle
        is counted twice (once for each direction). Note that self loops are not
        used to form triangles. 
        """

        A = self.adjacencyMatrix()
        A[numpy.diag_indices(A.shape[0])] = 0 
        A3 = numpy.linalg.matrix_power(A, 3)

        return numpy.array(numpy.diag(A3), numpy.int)

    def maxEigenvector(self):
        """
        Returns the eigenvector of maximum eigenvalue of the adjacency matrix. The
        eigenvector is of unit length, and measures the centrality of the corresponding
        vertex. It is based on the principle that connections to high-scoring nodes
        contribute more to the score of the node in question than equal connections
        to low-scoring nodes.
        """
        
        A = self.getWeightMatrix()
        w, V = numpy.linalg.eig(A)

        i = numpy.argmax(w)
        return V[:, i]

    def betweenness(self):
        """
        Return the betweenness of each vertex in the graph. The betweenness is
        defined as the number of shortest paths passing through each vertex. 
        """
        n = self.getNumVertices() 
        P = self.adjacencyMatrix().copy()
        P[P==0] = float('inf')
        N = numpy.ones((n, n), numpy.int32)*-1
        pathFreqs = numpy.zeros(n)

        for i in range(n):
            for j in range(n):
                for k in range(n):
                    if P[j, i] + P[i, k] <= P[j, k]:
                        P[j, k] = P[j,i]+P[i,k]
                        N[j, k] = i

        #How do we store next values for multiple shortest paths?
        print("jhjhj")
        print(pathFreqs)
        N2 = N[N!=-1]
        return numpy.bincount(N2.ravel())

    def clusteringCoefficient(self):
        """
        Find the clustering coefficient of this graph.
        """
        A = self.adjacencyMatrix()
        A2 = numpy.dot(A, A)
        A3 = numpy.dot(A2, A)

        diagSumA2 = 0

        for i in range(0, A2.shape[0]):
            diagSumA2 = diagSumA2 + A2[i, i]

        numTriples = A2.sum()-diagSumA2

        diagSumA3 = 0

        for i in range(0, A3.shape[0]):
            diagSumA3 = diagSumA3 + A3[i, i]

        if numTriples == 0:
            return 0
        else:
            return diagSumA3/numTriples

    def dijkstrasAlgorithm(self, vertexIndex):
        """
        Run Dijkstras Algorithm on the graph for a given source vertex. We ignore weights
        and simply use distance as the (unweighted) path length. Returns an array
        with the distance to all vertices (including itself). Method is not as fast
        as it could be due to accesses of csr matrix.

        :param vertexIndex: the index of the source vertex.
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())

        distance = numpy.ones(self.W.shape[0])*numpy.inf
        previous = numpy.ones(self.W.shape[0])*-1
        distance[vertexIndex] = 0

        A = self.adjacencyMatrix()
        #A = A.tocsc()
        #A = A.todense()
        vertexIndices = numpy.array(range(0, self.getNumVertices()))

        while vertexIndices.shape[0] != 0:
            minVertexIndex = scipy.argmin(distance[vertexIndices])
            minVertexDistance = distance[vertexIndices][minVertexIndex]

            if  minVertexDistance == numpy.inf:
                break

            #Arow = A.getrow(vertexIndices[minVertexIndex]).todense()
            if scipy.sparse.issparse(A): 
                Arow = A[vertexIndices[minVertexIndex], :].todense()
                Arow = numpy.ravel(Arow)
            else:
                Arow = A[vertexIndices[minVertexIndex], :]
                
            cols= numpy.nonzero(Arow)[0]

            for i in cols:
                newDistance = minVertexDistance + Arow[i]
                if newDistance < distance[i]:
                    distance[i] =  newDistance
                    previous[i] = minVertexIndex

            vertexIndices = numpy.delete(vertexIndices, minVertexIndex)

        #From previous we can reconstruct the path

        return distance
    
    def diameter2(self):
        """
        Use Dijkstras Algorithm to compute the diameter of the graph.
        """

        maxDiameter = 0

        for i in range(self.getNumVertices()):
            distances = self.dijkstrasAlgorithm(i)
            maxDistance = numpy.max(distances[distances!=numpy.inf])
            if maxDistance > maxDiameter:
                maxDiameter = maxDistance

        return maxDiameter
    
    def egoGraph(self, vertexIndex):
        """
        Returns the subgraph composed of the given vertex and its immediate neighbours.
        In the new graph, the ego is index 0 and neighbours are indexed in order
        after 0.

        :param vertexIndex: the index of the source vertex.
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())

        neighbours = self.neighbours(vertexIndex)
        egoGraphIndices = numpy.r_[numpy.array([vertexIndex]), neighbours]
        return self.subgraph(egoGraphIndices)

    def isTree(self):
        """
        Returns true if this graph is a tree. Every vertex must have an in-degree
        of 1 (i.e. one parent), except the root which has an in-degree of zero
        and non-zero out-degree.
        """
        if self.isUndirected():
            raise ValueError("Only directed graphs can be trees")

        inDegSeq = self.inDegreeSequence()
        j = numpy.ones(inDegSeq.shape[0])
        nonZeroInds = inDegSeq!=0
        zeroInds = inDegSeq==0

        if (inDegSeq[nonZeroInds] != j[nonZeroInds]).any():
            return False

        if numpy.sum(zeroInds) != 1:
            return False 

        #Find zero entry and ensure it has out-degree > 0
        root = numpy.nonzero(zeroInds)[0][0]
        neighbours = self.neighbours(root)

        if self.getNumVertices()>1 and neighbours.shape[0]==0:
            return False

        return True 

    def findTrees(self):
        """
        Returns a list of trees for a directed graph. The reason for only allowing
        directed graphs is that the root of a tree in an undirected graph is ambiguous.
        Each tree is represented by an list of indices of vertices in the graph.
        """
        if self.isUndirected():
            raise ValueError("Can only find trees on directed graphs")

        trees = []

        #Find all vertices with in-degree 0
        for i in range(0, self.getNumVertices()):
            if self.neighbourOf(i).shape[0] == 0:
                trees.append(self.depthFirstSearch(i))

        return trees

    vList = None
    undirected = None