import numpy
import logging
import sys
import cProfile
import pstats
import os
import scipy.sparse
from apgl.generator.ErdosRenyiGenerator import ErdosRenyiGenerator
from apgl.graph import *
from apgl.util import *

logging.basicConfig(stream=sys.stdout, level=logging.INFO)

outputDirectory = PathDefaults.getOutputDir()
directory = outputDirectory + "test/"
profileFileName = directory + "profile.cprof"

class SparseGraphProfile:
    def __init__(self):
        numVertices = 1000
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)
        W = scipy.sparse.lil_matrix((numVertices, numVertices))
        graph = SparseGraph(vList, W=W)
        generator = ErdosRenyiGenerator(graph)

        p = 0.4
        self.graph = generator.generateGraph(p)

        outputDirectory = PathDefaults.getOutputDir()
        directory = outputDirectory + "test/"
        self.profileFileName = directory + "profile.cprof"

    def profileSubgraph():
        numVertices = 500
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)
        graph = SparseGraph(vList)
        generator = ErdosRenyiGenerator(graph)

        p = 0.1

        graph = generator.generateGraph(p)
        indices = numpy.random.permutation(numVertices)

        cProfile.runctx('graph.getSubGraph(indices)', globals(), locals(), profileFileName)
        stats = pstats.Stats(profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)


    def profileDiameter():
        numVertices = 100
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)
        graph = SparseGraph(vList)
        generator = ErdosRenyiGenerator(graph)

        p = 0.1

        graph = generator.generateGraph(p)

        outputDirectory = PathDefaults.getOutputDir()
        directory = outputDirectory + "test/"
        profileFileName = directory + "profile.cprof"

        logging.info("Starting to profile ...")
        cProfile.runctx('graph.diameter2()', globals(), locals(), profileFileName)
        stats = pstats.Stats(profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)

    def profileDijkstrasAlgorithm(self):

        logging.info("Starting to profile ...")
        cProfile.runctx('self.graph.dijkstrasAlgorithm(0)', globals(), locals(), self.profileFileName)
        stats = pstats.Stats(self.profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)

    def profileFitPowerLaw():
        numVertices = 5000
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)
        graph = SparseGraph(vList, True)

        ell = 2
        m = 2
        generator = BarabasiAlbertGenerator(graph)
        generator.generateGraph(ell, m)
        #print(graph.degreeDistribution())

        alpha, ks, xmin = graph.fitPowerLaw()

        logging.info("Starting to profile ...")
        cProfile.runctx('graph.fitPowerLaw()', globals(), locals(), profileFileName)
        stats = pstats.Stats(profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)

    def profileSparseMatrices():
        A = numpy.random.random_integers(0, 1, (1000, 1000))
        #W = scipy.sparse.csr_matrix(A)
        W = scipy.sparse.coo_matrix(A)
        #W.sort_indices()

        #Results: lil_matrix has the fastest row operations but very slow column ones
        #csr_matrix has fast-ish row operations and slightly slower column ones
        #csc_matrix is the opposite

        def getRows():
            for i in range(W.shape[0]):
                W.getrow(i).getnnz()

            #W.tocsc()
            #for i in range(W.shape[0]):
            #    W.getcol(i).getnnz()

        logging.info("Starting to profile ...")
        cProfile.runctx('getRows()', globals(), locals(), profileFileName)
        stats = pstats.Stats(profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)

    def profileOutDegreeSequence(self):
        logging.info("Starting to profile ...")
        cProfile.runctx('self.graph.outDegreeSequence()', globals(), locals(), self.profileFileName)
        stats = pstats.Stats(self.profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)

    def profileNeighbours(self):
        def allNeighbours():
            for i in range(self.graph.getNumVertices()):
                self.graph.neighbours(i)

        logging.info("Starting to profile ...")
        cProfile.runctx('allNeighbours()', globals(), locals(), self.profileFileName)
        stats = pstats.Stats(self.profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)


    def profileAddEdges(self):
        """
        We want to try creating a huge graph and adding lots of edges.
        """
        numVertices = 1000000
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)
        W = scipy.sparse.lil_matrix((numVertices, numVertices))
        graph = SparseGraph(vList, W=W)

        def runAddEdges():
            numEdges = 100000
            edgesArray = numpy.zeros((numEdges, 2))

            for i in xrange(numEdges):
                vertexInd1 = numpy.random.randint(0, numVertices)
                vertexInd2 = numpy.random.randint(0, numVertices)
                edgesArray[i, 0] = vertexInd1
                edgesArray[i, 1] = vertexInd2

            graph.addEdges(edgesArray)

        logging.info("Starting to profile ...")
        cProfile.runctx('runAddEdges()', globals(), locals(), self.profileFileName)
        stats = pstats.Stats(self.profileFileName)
        stats.strip_dirs().sort_stats("cumulative").print_stats(20)
        

profiler = SparseGraphProfile()
profiler.profileAddEdges()
