import Util
import numpy
import bisect
from apgl.util.Util import *

"""
Model the contact rate of an infected individual and other susceptibles.
"""
class HIVRates():
    def __init__(self, graph):
        self.graph = graph
        self.degreeSequence = graph.outDegreeSequence()

        #Parameters for sexual contact
        self.alpha = 1.0
        self.newContactChance = 5.0
        self.heteroContactRate = 0.1
        self.biContactRate = 0.2

        #Infection probabilities 
        self.womanManInfectProb = 0.3
        self.manWomanHeteroInfectProb = 0.2
        self.manBiInfectProb = 0.5
        self.defaultInfectProb = 0.5

        #Random detection
        self.randDetectRate = 10.0

        #Contact tracing parameters s
        self.ctRatePerPerson = 0.05
        self.ctStartTime = 180
        self.ctEndTime = 730

        #Contact times is a list of neighbours of each vertex and the corresponding 
        #times of sexual contact.
        self.contactTimesList = []

        for i in range(graph.getNumVertices()):
            contactTimes = self.contactTimes(i)
            self.contactTimesList.append(contactTimes)

    def contactTimes(self, vertexInd):
        """
        Returns an array with 2 columns such that the 1st is the set of neighbours
        of the current vertex and the 2nd is the value on the edge (time of last
        sexual contact). 
        """
        neighbours = self.graph.neighbours(vertexInd)
        contactTimes = numpy.zeros((neighbours.shape[0], 2))

        for j in range(neighbours.shape[0]):
            contactTimes[j, 0] = neighbours[j]
            contactTimes[j, 1] = self.graph.getEdge(vertexInd, neighbours[j])

        return contactTimes

    def contactEvent(self, vertexInd1, vertexInd2, t):
        """
        Indicates a sexual contact event between two vertices. 
        """
        if self.graph.getEdge(vertexInd1, vertexInd2) == None:
            self.degreeSequence[vertexInd1] += 1
            self.degreeSequence[vertexInd2] += 1

        self.graph.addEdge(vertexInd1, vertexInd2, t)
        self.contactTimesList[vertexInd1] = self.contactTimes(vertexInd1)
        self.contactTimesList[vertexInd2] = self.contactTimes(vertexInd2)
            
    def contactRate(self, vertexInd1, susceptibleList, t):
        """
        Work out the contact rate. Note that susceptibleList must be sorted. 
        """
        if len(susceptibleList) == 0:
            return numpy.array([])
        
        contactRates = numpy.zeros(len(susceptibleList))

        #Preferential attachment - choose a random new contact from susceptibles
        #based on the degree. 
        contactTimes = self.contactTimesList[vertexInd1]
        newContactInd = susceptibleList[Util.randomChoice(self.degreeSequence[susceptibleList] + 1)]
        
        contactProbs = numpy.append((0.1+t-contactTimes[:, 1]), [self.newContactChance])**-self.alpha
        contactInds = numpy.append(contactTimes[:, 0], [newContactInd])

        #Now make choice based on contact probabilities
        vertexInd2 = contactInds[Util.randomChoice(contactProbs)]
        #NOTE: that we may choose someone who is not susceptible here

        if vertexInd2 in susceptibleList:
            susceptibleInd = bisect.bisect_left(susceptibleList, vertexInd2)
            #susceptibleInd = susceptibleList.index(vertexInd2)
            vertex1 = self.graph.getVertex(vertexInd1)
            vertex2 = self.graph.getVertex(vertexInd2)

            if vertex1.getFreeDegree() == 0 or vertex2.getFreeDegree() == 0:
                contactRates[susceptibleInd] = 0
            elif vertex1.getGender() != vertex2.getGender():
                contactRates[susceptibleInd] = self.heteroContactRate
            elif vertex1.getGender() == vertex2.getGender() and vertex1.getOrientation()==1 and vertex2.getOrientation()==1:
                contactRates[susceptibleInd] = self.biContactRate

        return contactRates


    def contactRates(self, infectedList, susceptibleList, t):
        """
        Work out contact rates between all infected and all susceptibles. What is
        computational complexity?
        """
        if len(infectedList) == 0 or len(susceptibleList):
            return numpy.array([])

        contactRates = sparse.scipy.lil_matrix((len(infectedList), len(susceptibleList)))
        
        #First choose a set of new contacts based on the degree sequence
        #These contacts can be any one of the other contacts
        possibleContacts = numpy.zeros((len(infectedList), 2))
        possibleContacts[:, 0] = self.oldContacts[infectedList, 0]
        possibleContacts[:, 1] = Util.randomChoice(self.degreeSequence+1, len(infectedList))

        possibleContactProbs = numpy.zeros((len(infectedList), 2))
        possibleContactProbs[:, 0] = self.oldContacts[infectedList, 1]
        possibleContactProbs[:, 1] = numpy.ones(len(infectedList))*self.newContactChance

        
        #Then make random choice
        #Then assign appropriate rate 

    """
    Compute the infection probability between an infected and susceptible.
    """
    def infectionProbability(self, vertexInd1, vertexInd2, t):
        vList = self.graph.getVertexList()
        vertex1 = vList.getVertex(vertexInd1)
        vertex2 = vList.getVertex(vertexInd2)

        #Gender 0 is female, 1 is male
        if vertex1.getGender() == 0 and vertex2.getGender() == 1:
            return self.womanManInfectProb
        if vertex1.getGender() == 1 and vertex2.getGender() == 0 and vertex1.getOrientation()==0:
            return self.manWomanHeteroInfectProb
        if vertex1.getGender() == 1 and vertex1.getOrientation()==1:
            return self.manBiInfectProb
        else:
            return self.defaultInfectProb

    """
    Compute the detection rate of an infected which depends on the entire population.
    """
    def randomDetectionRate(self, vertexInd, t):
        return self.randDetectRate/self.graph.getNumVertices()


    def contactTracingRate(self, vertexInd, t):
        """
        Compute the contact tracing detection rate of an infected.
        """
        #We can find the set of detected neigbours here
        neighbours = self.contactTimesList[vertexInd][:, 0]
        ctRate = 0.0

        #The rate is a sum of rates of detected contacts.
        for ind in neighbours:
            vertex = self.graph.getVertex(ind)
            timeSinceDetection = (t - vertex.getDetectionTime())

            if vertex.isRemoved() and timeSinceDetection >= self.ctStartTime and timeSinceDetection <= self.ctEndTime:
                ctRate += self.ctRatePerPerson

        return ctRate